/* This program is a part of the 'T_H_X10' project, found at
       http://members.home.net/mkeryan/t_h_x10/

    It is Copyrighted by under the GPL:  (c) 1999 Michael J. Keryan
    See the file "COPYING" included with this program.
    You are granted permission to use and modify the code with the
    stipulation that copyright notices remain and the author is to be
    notified of any improvements, bug fixes, etc.

   t_h_x10.c  This program inputs data in the form:
                 ADDR F01    or   FUNC F On   or   DATA 0x10
		 and processes it for temperature and humidity info.
	      It outputs data in the form:
	         H  32 F03 Date-Time-String
		 T 105 F01 Date-Time-String
*/

#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#define BUFSIZE 256
#define DIRPATH "/usr/local/t_h_x10/"

void newday(struct tm *timest);
void newmin(char date_time[BUFSIZE], time_t now, int min);
void newhr(int oldhr);
int parse_aline(char*, char*, char*, char*, int*);
void set_channel(int valid_index, time_t now);
void set_command(char valid_code[BUFSIZE], char date_time[BUFSIZE]);
void data_complete(int i, char date_time[BUFSIZE], int not_good);
int is_reasonable_T(int i, int the_data);
int is_reasonable_H(int i, int the_data);
void save_T(int i, int the_data);
void save_H(int i, int the_data);
int heat_index(int i);
void send_beep(void);

/* global variables */
char housecode[2];           /* e.g. 'F'                            */
char residence[BUFSIZE];     /* e.g. 'the Clinton residence'        */
char device[8][8];           /* e.g. "F01", "F07"                   */
char room[8][32];            /* e.g. "Bedroom_1", "Attic"           */
int zone[8];                 /* floor: 0 (outside), 1, 2, 3 (attic) */
int numloc;                  /* no. of possible sensors, e.g. 8     */
char beep[] = "\a";          /* beep for output                     */

struct fileptrs {
  FILE *warn;
  FILE *activ;
  FILE *data;
  FILE *avg;
  FILE *current;
  FILE *html;
} fileptrs = {NULL};

FILE *std_out = stdout;

struct run_avg {
  int num_T;
  int num_H;
  int T1, T2, T3, T4;
  int H1, H2, H3, H4;
  int T, H;
} run_avg[8] = {{0}};

struct avg {                 /* this data is flushed hourly      */
  int num_T;                 /* and averages saved to file       */
  int num_H;                 /* only data known to be good is    */
  int sum_T;                 /* put here, rejected if outside    */
  int sum_H;                 /* of running average, if an        */
  int avg_T;                 /* address is missing, or if a      */
  int avg_H;                 /* command is missing.              */
} avg[8] = {{0}};

struct current {
  int T;
  int H;
  int HI;
  int bits;                  /* temp. sum of input bits (T or H)  */
  int bit_count;             /* 1, 2, ... 128 for 8th bit rec'd   */
  int is_active;             /* 1 if active, 0 if not             */
  time_t start;              /* time first bit is received        */ 
} current[8] = {{0}};        /* compiler to initialize all to 0's */

int main(void)
{
  char astring[32];
  int numread, valid_index, valid_type;
  char aline[BUFSIZE], valid_code[BUFSIZE], valid_dev[BUFSIZE];
  char *aptr = NULL;
  char date_time[BUFSIZE];
  char std_out_str[BUFSIZE];
  struct tm *timest;
  time_t now;
  int day, min, hr;
  int oldday = -1, oldmin = -1, oldhr = -1;
  long flags;
  FILE *fileptr = NULL;

  /* read configuration data, first the housecode */
  if ((fileptr = fopen("/usr/local/t_h_x10/config/housecode","r")) == NULL) {
    fprintf(stderr,"Failed to open housecode file. Aborting\n");
    exit(-1);
  }
  if ((numread = fscanf(fileptr, "%1s", housecode)) < 1) {
    fprintf(stderr,"Failed to read housecode. Aborting\n");
    exit(-1);
  }

  /* now the redirected stdout (std_out) */
  if ((fileptr = fopen("/usr/local/t_h_x10/config/std_out","r")) == NULL) {
    fprintf(stderr,"Failed to open std_out; defaulting to stdout\n");
  } else {
    if ((fgets(std_out_str, sizeof(std_out_str), fileptr)) == NULL) {
      fprintf(stderr,"Failed to read std_out; defaulting to stdout\n");
    } else {
      fprintf(stdout,"Opening %s\n", std_out_str);
      if ((fileptr = fopen(std_out_str,"w")) == NULL) {
	fprintf(stderr,"Opening of %s failed, using stdout\n", std_out_str);
      } else {
	std_out = fileptr;
	if (fwrite(beep, 1, 1, std_out) != 1) {
	  fprintf(stderr,"Problem outputting beep to std_out\n");
	  fprintf(stderr,"Defaulting to stdout\n");
	  std_out = stdout;
	} else {
	  fprintf(stderr,"Output will be directed to %s\n", std_out_str);
	  fprintf(std_out,"Output is being directed to %s\n", std_out_str);
	}
      }
    }
  }

  /* residence string for html output */
  if ((fileptr = fopen("/usr/local/t_h_x10/config/residence","r")) == NULL) {
    strcpy(residence,"Copyright 1999 m.j.keryan\n");
  } else {  
    if ((fgets(residence, sizeof(residence), fileptr)) == NULL) {
      strcpy(residence,"Copyright 1999 m.j.keryan\n");
    }
  }

  /* locations of sensors */
  if ((fileptr = fopen("/usr/local/t_h_x10/config/locations","r")) == NULL) {
    fprintf(stderr,"Failed to open locations file. Aborting\n");
    exit(-1);
  }

  numloc = 0;
  while ((numloc < 8) && 
	 ((numread = fscanf(fileptr, "%2s %d %s", astring,
			    &zone[numloc], room[numloc])) > 2)) {
    strcpy(device[numloc],housecode);
    strcat(device[numloc],astring);
    numloc++;
  }

  /* set stdin to nonblocking */
  if ((flags = fcntl(STDIN_FILENO, F_GETFL)) == -1) {
    fprintf(stderr,"Failed to get STDIN flags. Aborting.\n");
    exit(-1);
  }
  flags |= O_NONBLOCK;
  if ((fcntl(STDIN_FILENO, F_SETFL, flags)) == -1) {
    fprintf(stderr,"Failed to change STDIN flags. Aborting.\n");
    exit(-1);
  }
  
  /* loop forever as below */
  for (;;) {
    now = time(NULL);
    if ((timest = localtime(&now)) == NULL) fprintf(stderr, "Time error!\n");
    day = timest->tm_yday;
    min = timest->tm_min;
    hr  = timest->tm_hour;
    strftime(date_time, BUFSIZE-1, "%I:%M:%S %p", timest);

    if (hr  != oldhr) {
      newhr(oldhr);
      oldhr = hr;
    }
    if (day != oldday) {
      newday(timest);
      oldday = day;
    }
    if (min != oldmin) {
      newmin(date_time, now, min);
      oldmin = min;
    }
    if ((aptr = fgets(aline, sizeof(aline), stdin)) == NULL) {
      sleep(1);
    } else {
      valid_type = parse_aline(aline, date_time, 
			       valid_code, valid_dev, &valid_index);
      if (valid_type == 1) {            /* command info, e.g. F On */
	set_command(valid_code, date_time);
      } else if (valid_type == 2) {     /* channel info, e.g. F01  */
	set_channel(valid_index, now);
      }
    }
  }
  return 0;
}

/* each day, open up new log files */
void newday(struct tm *timest) {

  char datestr[BUFSIZE];
  char filename[BUFSIZE];

  /* close old files if open before opening new ones */
  if (fileptrs.warn    != NULL) {
    if (fclose(fileptrs.warn) == EOF) {
      fprintf(stderr,"Failed to close warning log file. Aborting\n");
      exit(-1);
    };
  }
  if (fileptrs.activ   != NULL) {
    if (fclose(fileptrs.activ) == EOF) {
      fprintf(stderr,"Failed to close activity log file. Aborting\n");
      exit(-1);
    };
  }
  if (fileptrs.data    != NULL) {
    if (fclose(fileptrs.data) == EOF) {
      fprintf(stderr,"Failed to close data log file. Aborting\n");
      exit(-1);
    };
  }
  if (fileptrs.avg     != NULL) {
    if (fclose(fileptrs.avg) == EOF) {
      fprintf(stderr,"Failed to close data averages file. Aborting\n");
      exit(-1);
    };
  }

  /* open /usr/local/t_h_x10/logs/warn_yymmdd.log file */
  strftime(datestr,BUFSIZE-1,"%Y%m%d",timest);
  strcpy(filename,DIRPATH);
  strcat(filename,"logs/warn_");
  strcat(filename,datestr);
  strcat(filename,".log");
  if ((fileptrs.warn = fopen(filename,"a")) == NULL) {
    fprintf(stderr,"Failed to open warning log file. Aborting\n");
    exit(-1);
  }

  /* open /usr/local/t_h_x10/logs/activ_yymmdd.log file */
  strcpy(filename,DIRPATH);
  strcat(filename,"logs/activ_");
  strcat(filename,datestr);
  strcat(filename,".log");
  if ((fileptrs.activ = fopen(filename,"a")) == NULL) {
    fprintf(stderr,"Failed to open activity log file. Aborting\n");
    exit(-1);
  }

  /* open /usr/local/t_h_x10/data/data_yymmdd.log file */
  strcpy(filename,DIRPATH);
  strcat(filename,"data/data_");
  strcat(filename,datestr);
  strcat(filename,".log");
  if ((fileptrs.data = fopen(filename,"a")) == NULL) {
    fprintf(stderr,"Failed to open data log file. Aborting\n");
    exit(-1);
  }

  /* open /usr/local/t_h_x10/data/avg_yymmdd.log file */
  strcpy(filename,DIRPATH);
  strcat(filename,"data/avg_");
  strcat(filename,datestr);
  strcat(filename,".log");
  if ((fileptrs.avg = fopen(filename,"a")) == NULL) {
    fprintf(stderr,"Failed to open data averages file. Aborting\n");
    exit(-1);
  }
}

/* each hour, print averages for the data */
void newhr(int oldhr) {

  int i, z, flr_t[8], flr_h[8], flrcnt_t[8], flrcnt_h[8], flr_t_avg, flr_h_avg;

  for (i = 0; i < 8; i++) {
    flr_t[i] = 0;
    flr_h[i] = 0;
    flrcnt_t[i] = 0;
    flrcnt_h[i] = 0;
  }

  for (i = 0; i < numloc; i++) {
    if ((avg[i].num_T > 0) || (avg[i].num_H > 0)) {
      if (oldhr > -1) {
	/* write averages to file */
	fprintf(fileptrs.avg, "Hr= %2d %s  T=%4d n= %d  H=%4d n= %d\n",
		oldhr, device[i], avg[i].avg_T, avg[i].num_T,
		avg[i].avg_H, avg[i].num_H);

	/* sum the data for each zone (floor) */
	z = zone[i];
	if (avg[i].num_T > 0) {
	  flr_t[z] = flr_t[z] + avg[i].avg_T;
	  flrcnt_t[z] = flrcnt_t[z] + 1;
	}
	if (avg[i].num_H > 0) {
	  flr_h[z] = flr_h[z] + avg[i].avg_H;
	  flrcnt_h[z] = flrcnt_h[z] + 1;
	}
      }
      
      /* if no data, reset running average and current*/
      if (avg[i].num_T < 1) {
	run_avg[i].num_T = 0;
	current[i].T = 0;
	current[i].HI = 0;
      }

      if (avg[i].num_H < 1) {
	run_avg[i].num_H = 0;
	current[i].H = 0;
	current[i].HI = 0;
      }

      /* reset the hourly data */
      avg[i].num_T = 0;
      avg[i].sum_T = 0;
      avg[i].avg_T = 0;
      avg[i].num_H = 0;
      avg[i].sum_H = 0;
      avg[i].avg_H = 0;
    }
  }

  /* print zone averages for any zones with > 1 channel */
  for (i = 0; i < 8; i++) {
    if ((flrcnt_t[i] > 1) || (flrcnt_h[i] > 1)) {
      if (flrcnt_t[i] > 0) {
	flr_t_avg = flr_t[i] / flrcnt_t[i];
      } else { flr_t_avg = 0; }
      if (flrcnt_h[i] > 0) {
	flr_h_avg = flr_h[i] / flrcnt_h[i];
      } else { flr_h_avg = 0; }
      fprintf(fileptrs.avg, "Hr= %2d Z0%d  T=%4d n= %d  H=%4d n= %d\n",
	      oldhr, i, flr_t_avg, flrcnt_t[i], flr_h_avg, flrcnt_h[i]);
    }
  }

  fflush(fileptrs.avg);
}

/* each minute, check for timeouts */
void newmin(char date_time[BUFSIZE], time_t now, int min) {

  char dirname[BUFSIZE], filename1[BUFSIZE], filename2[BUFSIZE];
  int i;

  for (i=0; i < numloc; i++) {
    if ((current[i].start != 0) &&
	(difftime(now, current[i].start) > 120)) {
      fprintf(fileptrs.warn, "%s Timeout of channel: %s, data=%d\n",
	      date_time, device[i], current[i].bits);
      current[i].start = 0;
      current[i].bits = 0;
      current[i].bit_count = 0;
      current[i].is_active = 0;
    }
  }

  /* also write current data to file */
  strcpy(dirname,DIRPATH);
  strcpy(filename1,dirname);
  strcat(filename1,"data/current.temp");
  if ((fileptrs.current = fopen(filename1,"w")) == NULL) {
    fprintf(stderr,"Failed to open current data file.\n");
  } else {
    fprintf(fileptrs.current, "%15s      Temp %%Hum H.I.\n", date_time);
    for (i = 0; i < numloc; i++) {
      fprintf(fileptrs.current, "%3s %-16s %4d %4d %4d\n",
	      device[i], room[i], current[i].T, current[i].H, current[i].HI);
    }
    fclose(fileptrs.current);
    strcpy(filename2,dirname);
    strcat(filename2, "data/current.dat");
    rename(filename1, filename2);
  }

  /* write current data to html file, for use with browsers */
  strcpy(dirname,DIRPATH);
  strcpy(filename1,dirname);
  strcat(filename1,"data/temp.html");
  if ((fileptrs.html = fopen(filename1,"w")) == NULL) {
    fprintf(stderr,"Failed to open current data file.\n");
  } else {
     fprintf(fileptrs.html, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n");
     fprintf(fileptrs.html, "<HTML>\n");
     fprintf(fileptrs.html, "<HEAD>\n");
     fprintf(fileptrs.html, "<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=iso-8859-1\">\n");
     fprintf(fileptrs.html, "<TITLE>T_H_X10 T/H Report</TITLE>\n");
     fprintf(fileptrs.html, "<META NAME=\"GENERATOR\" CONTENT=\"t_h_x10\">\n");
     fprintf(fileptrs.html, "<META HTTP-EQUIV=\"REFRESH\" CONTENT=\"60\">\n");
     fprintf(fileptrs.html, "</HEAD>\n");
     fprintf(fileptrs.html, "<BODY TEXT=\"#000000\" BGCOLOR=\"#006600\" LINK=\"#0000EF\" VLINK=\"#52188C\" ALINK=\"#FF0000\">\n");
     fprintf(fileptrs.html, "<HR WIDTH=\"100%%\">\n");
     fprintf(fileptrs.html, "<CENTER><H1>\n");
     fprintf(fileptrs.html, "<B><FONT COLOR=\"#FFFFFF\">Temperature and Humidity Monitor -- %s</FONT></B></H1></CENTER>\n", residence);
     fprintf(fileptrs.html, "<HR WIDTH=\"100%%\"></CENTER>\n");
     fprintf(fileptrs.html, "<CENTER><H2>\n");
     fprintf(fileptrs.html, "<FONT COLOR=\"#FFFF00\">Time:&nbsp; %s</FONT></H2></CENTER>\n", date_time);
     fprintf(fileptrs.html, "<CENTER><TABLE BORDER=2 CELLSPACING=2 CELLPADDING=2 WIDTH=\"80%%\" BGCOLOR=\"#FFFFCC\" NOSAVE >\n");
     fprintf(fileptrs.html, "<TR ALIGN=CENTER VALIGN=CENTER NOSAVE>\n");
     fprintf(fileptrs.html, "<TD>Channel #</TD>\n");
     fprintf(fileptrs.html, "<TD>Location</TD>\n");
     fprintf(fileptrs.html, "<TD>Temp. -- deg F</TD>\n");
     fprintf(fileptrs.html, "<TD>%% Humidity</TD>\n");
     fprintf(fileptrs.html, "<TD>Heat Index</TD>\n");
     fprintf(fileptrs.html, "</TR>\n");
     for (i = 0; i < numloc; i++) {
       fprintf(fileptrs.html, "<TR ALIGN=CENTER VALIGN=CENTER NOSAVE>\n");
       fprintf(fileptrs.html, "<TD>%s</TD>\n", device[i]);
       fprintf(fileptrs.html, "<TD>%s</TD>\n", room[i]);
       fprintf(fileptrs.html, "<TD>%4d</TD>\n",current[i].T);
       fprintf(fileptrs.html, "<TD>%4d</TD>\n",current[i].H);
       fprintf(fileptrs.html, "<TD>%4d</TD>\n",current[i].HI);
       fprintf(fileptrs.html, "</TR>\n");
     }
     fprintf(fileptrs.html, "</TABLE></CENTER>\n");
     fprintf(fileptrs.html, "<BR>&nbsp;\n");
     fprintf(fileptrs.html, "<P><HR WIDTH=\"100%%\">\n");
     fprintf(fileptrs.html, "<CENTER><img SRC=\"/mnt/t_h_x10/data/plot.gif\" height=480 width=640></CENTER>\n");
     fprintf(fileptrs.html, "</BODY>\n");
     fprintf(fileptrs.html, "</HTML>\n");
     fclose(fileptrs.html);
     strcpy(filename2,dirname);
     strcat(filename2, "data/current.html");
     rename(filename1, filename2);
  }
  fflush(fileptrs.warn);
  fflush(fileptrs.current);
  fflush(fileptrs.html);

  /* now send a beep */
  send_beep();
}

int parse_aline(char aline[BUFSIZE], char date_time[BUFSIZE],
		char valid_code[BUFSIZE], char valid_dev[BUFSIZE],
		int *valid_index) {
  
  char input_type[BUFSIZE], input_channel[BUFSIZE], input_command[BUFSIZE];
  int numargs, i, this_channel;
  int valid_type = 0;

  /* parse the line */
  numargs = sscanf(aline, "%4s %s %s", 
		   input_type, input_channel, input_command);
  if (numargs < 2)
    fprintf(fileptrs.warn, "%s Bad data received  : %s\n", date_time, aline);
  else {
    if (strcmp(input_type, "ADDR") == 0) {         /* an address */
      this_channel = -1;
      
      for (i=0; i < numloc; i++)
	if (strcmp(input_channel, device[i]) == 0) this_channel = i;
      if (this_channel == -1) 
	fprintf (fileptrs.activ, "%s  Unmonitored address: %s\n",
		 date_time, input_channel);
      else {
	strcpy(valid_dev, device[this_channel]);
	*valid_index = this_channel;
	valid_type = 2;
      }
    } else if ((strcmp(input_type, "FUNC") == 0) && (numargs > 2)) {
      /* a function */
      if (strcmp(input_channel, housecode) != 0) {
	fprintf (fileptrs.activ, "%s  Unmonitored code   : %s %s\n",
		 date_time, input_channel, input_command);
      } else {   /* they're equal */
	strcpy(valid_code, input_command);
	valid_type = 1;
      }
    } else if (strcmp(input_type, "DATA") == 0) {  /* a data byte */
      fprintf (fileptrs.activ, "%s  Unmonitored data   : %s",
	       date_time, aline);
    } else {                                       /* unknown type */
      fprintf (fileptrs.warn, "%s  Unknown data type  : %s", date_time, aline);
    }
    
  }
  fflush(fileptrs.warn);
  fflush(fileptrs.activ);
  return(valid_type);
}

void set_channel(int valid_index, time_t now) {

  if (current[valid_index].start == 0) {
    current[valid_index].start = now;
    current[valid_index].is_active = 1;
  } else {
    current[valid_index].is_active += 1;
  }
}

void set_command(char valid_code[BUFSIZE], char date_time[BUFSIZE]) {

  static int last_index = -1;
  int abit = -1, not_good = -1;
  int at_least_one = -1;
  int i, j;

  if (strcmp(valid_code, "On") == 0) {
    abit = 1;
  } else if (strcmp(valid_code, "Off") == 0) {
    abit = 0;
  } else {
    fprintf(fileptrs.warn, "%s Invalid command: %s\n",
	      date_time, valid_code);
  }

  if (abit >= 0) {
    /* if 'F4' 'F On' 'F On' sequence, assume F4 for 2nd command also */ 
    for (i = 0; i < numloc; i++) {
      if (current[i].is_active > 0) {
	at_least_one = 1;
	break;
      }
    }
    if (at_least_one < 0) {
// disable guessing
//      current[last_index].is_active = 1;
      not_good = 1;
      fprintf(fileptrs.warn, "%s Missing address, assuming: %s\n",
	      date_time, device[last_index]);
    }
      
    for (i = 0; i < numloc; i++) {
      if (current[i].is_active > 0) {
	for (j = 1; j <= current[i].is_active; j++) {
	  if (current[i].bit_count == 0) {
	    current[i].bit_count = 1;
	  } else {
	    current[i].bit_count *= 2;
	  }
	  if (current[i].bit_count > 255) {
	    current[i].bits = 0;
	    current[i].bit_count = 0;
	    current[i].start = 0;
	    fprintf(fileptrs.warn, "%s More than 8 bits: %s\n",
		    date_time, device[i]);
	  } else {           /* data is ok, build up byte */
	    current[i].bits += (current[i].bit_count * abit);
	    last_index = i;
	    if (j > 1) {
	      not_good = 2;
	      fprintf(fileptrs.warn, "%s Missing command, assuming: %s\n",
		      date_time, valid_code);
// disable guessing 
              current[i].bit_count = 0;
              current[i].is_active = 0;
	    }
	    if (current[i].bit_count > 127) {
	      data_complete(i, date_time, not_good);
	    }
	  }
	}
	current[i].is_active = 0;
      }
    }
  }
  fflush(fileptrs.warn);
}

void data_complete(int i, char date_time[BUFSIZE], int not_good) {
  
  int the_data;
  char data_type;
  char message[8];

  if (not_good < 0) {
    strcpy(message, "OK");
  } else if (not_good == 1) {
    strcpy(message, "ADR-ERR");
  } else if (not_good == 2) {
    strcpy(message, "COM-ERR");
  } else {
    strcpy(message, "OTH-ERR");
  }

  if (current[i].bits > 127) {
    data_type = 'H';
    the_data = current[i].bits - 128;
    if (is_reasonable_H(i, the_data)) {
      if (not_good < 1) {
	save_H(i, the_data);
      }
    } else {
      strcpy(message, "HUM-ERR");
    }
  } else {
    data_type = 'T';
    the_data = current[i].bits;
    if (is_reasonable_T(i, the_data)) {
      if (not_good < 1) {
	save_T(i, the_data);
      }
    } else {
      strcpy(message, "TEM-ERR");
    }
  }
  
  current[i].bits = 0;
  current[i].bit_count = 0;
  current[i].start = 0;
  fprintf(std_out, "%c %3d %s %s %s\n",
	 data_type, the_data, device[i], date_time, message);
  fprintf(fileptrs.data, "%c %3d %s %s %s\n", 
	  data_type, the_data, device[i], date_time, message);
  fflush(std_out);
  fflush(fileptrs.data);
}

void save_H(int i, int the_data) {

  current[i].H = the_data;
  current[i].HI = heat_index(i);
  avg[i].num_H += 1;
  avg[i].sum_H += the_data;
  avg[i].avg_H = avg[i].sum_H / avg[i].num_H;
}

void save_T(int i, int the_data) {

  current[i].T = the_data;
  current[i].HI = heat_index(i);
  avg[i].num_T += 1;
  avg[i].sum_T += the_data;
  avg[i].avg_T = avg[i].sum_T / avg[i].num_T;
}

int is_reasonable_T(int i, int the_data) {
  
  int allowable_diff = 15;
  int is_ok = 1;

  if (run_avg[i].num_T < 1) {
    run_avg[i].num_T = 1;
    run_avg[i].T1 = the_data;
    run_avg[i].T = the_data;
  } else if (((abs(the_data - run_avg[i].T)) < allowable_diff) ||
	     (run_avg[i].num_T < 4)) {
    run_avg[i].T4 = run_avg[i].T3;
    run_avg[i].T3 = run_avg[i].T2;
    run_avg[i].T2 = run_avg[i].T1;
    run_avg[i].T1 = the_data;
    if (run_avg[i].num_T < 4) {
      run_avg[i].num_T += 1;
    }
    run_avg[i].T = (run_avg[i].T1 + run_avg[i].T2 + run_avg[i].T3
		    + run_avg[i].T4) / run_avg[i].num_T;
  } else {
    is_ok = 0;
  }
  return (is_ok);
}

int is_reasonable_H(int i, int the_data) {
  
  int allowable_diff = 25;
  int is_ok = 1;

  if (run_avg[i].num_H < 1) {
    run_avg[i].num_H = 1;
    run_avg[i].H1 = the_data;
    run_avg[i].H = the_data;
  } else if (((abs(the_data - run_avg[i].H)) < allowable_diff) ||
	     (run_avg[i].num_H < 4)) {
    run_avg[i].H4 = run_avg[i].H3;
    run_avg[i].H3 = run_avg[i].H2;
    run_avg[i].H2 = run_avg[i].H1;
    run_avg[i].H1 = the_data;
    if (run_avg[i].num_H < 4) {
      run_avg[i].num_H += 1;
    }
    run_avg[i].H = (run_avg[i].H1 + run_avg[i].H2 + run_avg[i].H3
		    + run_avg[i].H4) / run_avg[i].num_H;
  } else {
    is_ok = 0;
  }
  return (is_ok);
}

int heat_index(i) {

  double T, H, HI;
  int int_hi;
  /* from:  http://taiga.geog.niu.edu/nwslot/hiwc.html#heat_index */
  /* formula not used for low temperatures  (< 75 F) */
  if ((current[i].H == 0) || (current[i].T < 75)) {
    int_hi = current[i].T;
  } else {
    H = current[i].H;
    T = current[i].T;
    HI = 16.923 + (0.185212 * T) + (5.37941 * H) - (0.100254 * T * H);
    HI = HI + (0.00941695 * T * T) + (0.00728898 * H * H);
    HI = HI + (0.000345372 * T * T * H);
    HI = HI - (0.000814971 * T * H * H);
    HI = HI + (0.0000102102 * T * T * H * H);
    HI = HI - (0.000038648 * T * T * T);
    HI = HI + (0.0000291583 * H * H * H);
    HI = HI + (0.00000142721 * T * T * T * H);
    HI = HI + (0.000000197483 * T * H * H * H);
    HI = HI - (0.0000000218429 * T * T * T * H * H);
    HI = HI + (0.000000000843296 * T * T * H * H * H);
    HI = HI - (0.0000000000481975 * T * T * T * H * H * H);
    int_hi = HI;
  }
  current[i].HI = int_hi;
  return int_hi;
}

void send_beep(void) {

  if (fwrite(beep, 1, 1, std_out) != 1) {
    fprintf(stderr,"Problem outputting beep to std_out\n");
  }
  fflush(std_out);
}
